// Copyright 2022 Dennis Hezel
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef AGRPC_DETAIL_CONDITIONAL_SENDER_HPP
#define AGRPC_DETAIL_CONDITIONAL_SENDER_HPP

#include <agrpc/detail/asio_forward.hpp>
#include <agrpc/detail/config.hpp>
#include <agrpc/detail/forward.hpp>
#include <agrpc/detail/receiver.hpp>
#include <agrpc/detail/tuple.hpp>
#include <agrpc/detail/utility.hpp>

AGRPC_NAMESPACE_BEGIN()

namespace detail
{
template <class Sender, class... CompletionArgs>
class ConditionalSender;

struct ConditionalSenderAccess
{
    template <class Sender, class... CompletionArgs>
    static auto create(Sender&& sender, bool condition, CompletionArgs&&... args)
    {
        return detail::ConditionalSender(static_cast<Sender&&>(sender), condition,
                                         static_cast<CompletionArgs&&>(args)...);
    }
};

template <class Sender, class... CompletionArgs>
class ConditionalSender
{
  private:
    using ArgsTuple = detail::Tuple<CompletionArgs...>;

  public:
    template <template <class...> class Variant, template <class...> class Tuple>
    using value_types = typename Sender::template value_types<Variant, Tuple>;

    template <template <class...> class Variant>
    using error_types = typename Sender::template error_types<Variant>;

    static constexpr bool sends_done = Sender::sends_done;

    template <class Receiver>
    detail::ConditionalSenderOperationState<Sender, detail::RemoveCrefT<Receiver>, CompletionArgs...> connect(
        Receiver&& receiver) && noexcept((detail::IS_NOTRHOW_DECAY_CONSTRUCTIBLE_V<Receiver> &&
                                          std::is_nothrow_move_constructible_v<Sender>))
    {
        return {static_cast<Receiver&&>(receiver), static_cast<Sender&&>(sender_), condition_,
                static_cast<ArgsTuple&&>(args_)};
    }

    template <class Receiver, class S = Sender, class = std::enable_if_t<std::is_copy_constructible_v<S>>>
    detail::ConditionalSenderOperationState<Sender, detail::RemoveCrefT<Receiver>, CompletionArgs...> connect(
        Receiver&& receiver) const& noexcept((detail::IS_NOTRHOW_DECAY_CONSTRUCTIBLE_V<Receiver> &&
                                              std::is_nothrow_copy_constructible_v<Sender>))
    {
        return {static_cast<Receiver&&>(receiver), sender_, condition_, args_};
    }

  private:
    friend detail::ConditionalSenderAccess;

    ConditionalSender(Sender&& sender, bool condition, CompletionArgs&&... args)
        : sender_{static_cast<Sender&&>(sender)}, args_{static_cast<CompletionArgs&&>(args)...}, condition_(condition)
    {
    }

    Sender sender_;
    ArgsTuple args_;
    bool condition_;
};

template <class Sender, class... CompletionArgs>
ConditionalSender(Sender, bool, CompletionArgs...) -> ConditionalSender<Sender, CompletionArgs...>;

template <class Variant>
struct ConditionalSenderSatisfyReceiver;

template <class... T, class... Tuple>
struct ConditionalSenderSatisfyReceiver<detail::TypeList<detail::TypeList<T...>, Tuple...>>
{
    template <class Receiver, class... Args, size_t... I>
    static void satisfy_impl(Receiver&& receiver, detail::Tuple<Args...>&& tuple, std::index_sequence<I...>)
    {
        detail::satisfy_receiver(static_cast<Receiver&&>(receiver),
                                 detail::get<I>(static_cast<detail::Tuple<Args...>&&>(tuple))...);
    }

    template <class Receiver, class... Args>
    static void satisfy(Receiver&& receiver, detail::Tuple<Args...>&& tuple)
    {
        ConditionalSenderSatisfyReceiver::satisfy_impl(static_cast<Receiver&&>(receiver),
                                                       static_cast<detail::Tuple<Args...>&&>(tuple),
                                                       std::make_index_sequence<sizeof...(Args)>{});
    }

    template <class Receiver>
    static void satisfy(Receiver&& receiver, detail::Tuple<>&&)
    {
        detail::satisfy_receiver(static_cast<Receiver&&>(receiver), T{}...);
    }
};

template <class Sender, class Receiver, class... CompletionArgs>
class ConditionalSenderOperationState
{
  private:
    using ArgsTuple = detail::Tuple<CompletionArgs...>;

  public:
    void start() noexcept
    {
        if (condition_)
        {
            detail::exec::start(operation_state_);
        }
        else
        {
            using CompletionValues = typename Sender::template value_types<detail::TypeList, detail::TypeList>;
            detail::ConditionalSenderSatisfyReceiver<CompletionValues>::satisfy(
                static_cast<Receiver&&>(operation_state_.receiver()), static_cast<ArgsTuple&&>(args_));
        }
    }

  private:
    friend detail::ConditionalSender<Sender, CompletionArgs...>;

    template <class R>
    ConditionalSenderOperationState(R&& receiver, Sender&& sender, bool condition, ArgsTuple&& args)
        : operation_state_(detail::exec::connect(static_cast<Sender&&>(sender), static_cast<R&&>(receiver))),
          args_(static_cast<ArgsTuple&&>(args)),
          condition_(condition)
    {
    }

    template <class R>
    ConditionalSenderOperationState(R&& receiver, const Sender& sender, bool condition, const ArgsTuple& args)
        : operation_state_(detail::exec::connect(sender, static_cast<R&&>(receiver))),
          args_(args),
          condition_(condition)
    {
    }

    detail::exec::connect_result_t<Sender, Receiver> operation_state_;
    ArgsTuple args_;
    bool condition_;
};
}

AGRPC_NAMESPACE_END

#ifdef AGRPC_ASIO_HAS_SENDER_RECEIVER
#if !defined(BOOST_ASIO_HAS_DEDUCED_CONNECT_MEMBER_TRAIT) && !defined(ASIO_HAS_DEDUCED_CONNECT_MEMBER_TRAIT)
template <class Sender, class Receiver, class... CompletionArgs>
struct agrpc::asio::traits::connect_member<agrpc::detail::ConditionalSender<Sender, CompletionArgs...>, Receiver>
{
    static constexpr bool is_valid = true;
    static constexpr bool is_noexcept = noexcept(
        std::declval<agrpc::detail::ConditionalSender<Sender, CompletionArgs...>>().connect(std::declval<Receiver>()));

    using result_type = decltype(std::declval<agrpc::detail::ConditionalSender<Sender, CompletionArgs...>>().connect(
        std::declval<Receiver>()));
};
#endif

#if !defined(BOOST_ASIO_HAS_DEDUCED_START_MEMBER_TRAIT) && !defined(ASIO_HAS_DEDUCED_START_MEMBER_TRAIT)
template <class Sender, class Receiver, class... CompletionArgs>
struct agrpc::asio::traits::start_member<
    agrpc::detail::ConditionalSenderOperationState<Sender, Receiver, CompletionArgs...>>
{
    static constexpr bool is_valid = true;
    static constexpr bool is_noexcept = true;

    using result_type = void;
};
#endif
#endif

#endif  // AGRPC_DETAIL_CONDITIONAL_SENDER_HPP
